from pathlib import Path
from typing import Union

from .ProjectInstance import Activity, Mode, Project, ProjectInstance, Resource


def parse_mplib(loc: Union[str, Path]) -> ProjectInstance:
    """
    Parses a multi-project resource-constrained project scheduling problem
    instances from MPLIB.

    Parameters
    ----------
    loc
        The location of the instance.

    Returns
    -------
    Instance
        The parsed instance.
    """
    with open(loc, "r") as fh:
        lines = iter(line.strip() for line in fh.readlines() if line.strip())

    num_projects = int(next(lines))
    num_resources = int(next(lines))

    capacities = list(map(int, next(lines).split()))
    resources = [Resource(capacity=cap, renewable=True) for cap in capacities]

    projects: list[Project] = []
    activities: list[Activity] = []
    id2idx: dict[str, int] = {}  # maps activity names to idcs

    for project_idx in range(1, num_projects + 1):
        num_activities, release_date = map(int, next(lines).split())
        next(lines)  # denotes used resources, implies that demand > 0

        idcs = [len(activities) + idx for idx in range(num_activities)]
        projects.append(Project(idcs, release_date))

        for activity_idx in range(1, num_activities + 1):
            line = next(lines).split()
            idx = num_resources + 2
            duration, *demands, num_successors = list(map(int, line[:idx]))
            successors = line[idx:]
            assert len(successors) == num_successors

            mode = Mode(duration, demands)
            name = f"{project_idx}:{activity_idx}"  # original activity id
            id2idx[name] = len(activities)
            activities.append(Activity([mode], successors, name=name))  # type: ignore

    for activity in activities:
        # Map the successors ids from {project_idx:activity_idx}, to the
        # specific activitiy indices.
        idcs = [id2idx[succ] for succ in activity.successors]  # type: ignore
        activity.successors = idcs

    return ProjectInstance(resources, activities, projects)



import os


def relative(loc):
    """
    Returns the absolute path to the given location relative to this file.
    """
    current_dir = os.path.dirname(os.path.abspath(__file__))
    return os.path.join(current_dir, loc)



from pathlib import Path
from typing import Union

from .ProjectInstance import Activity, Mode, Project, ProjectInstance, Resource


def parse_rcpsp_ps(instance_loc: Union[str, Path]) -> ProjectInstance:
    """
    Parses a RCPSP-PS formatted instance from Van der Beek et al. (2024).
    """
    with open(instance_loc, "r") as fh:
        lines = iter(line.strip() for line in fh.readlines() if line.strip())

    num_activities, num_renewable, _ = map(int, next(lines).split())
    capacities = list(map(int, next(lines).split()))

    resources = [
        Resource(capacity, idx < num_renewable)  # resources are ordered
        for idx, capacity in enumerate(capacities)
    ]
    activities = []

    for idx in range(num_activities):
        duration, *demands = map(int, next(lines).split())
        line = map(int, next(lines).split())

        groups = []
        num_groups = next(line)
        for _ in range(num_groups):
            num_successors = next(line)
            groups.append([next(line) for _ in range(num_successors)])

        num_successors, *successors = map(int, next(lines).split())
        activities.append(
            Activity(
                [Mode(duration, demands)],
                successors,
                optional=idx > 0,  # source activity is not optional
                selection_groups=groups,
            )
        )

    project = Project(list(range(num_activities)))
    return ProjectInstance(resources, activities, [project])



from numpy.testing import assert_equal

from psplib import parse_psplib

from .utils import relative


def test_instance_single_mode():
    """
    Tests an PSPLIB instance with a single mode (aka RCPSP).
    """
    instance = parse_psplib(relative("data/m11_1.mm"))

    capacities = [res.capacity for res in instance.resources]
    renewables = [res.renewable for res in instance.resources]

    assert_equal(instance.num_resources, 4)
    assert_equal(capacities, [12, 9, 37, 53])
    assert_equal(renewables, [True, True, False, False])

    assert_equal(instance.num_activities, 18)

    activity = instance.activities[1]  # second activity (jobnr. 2)
    successors = [4, 8]
    assert_equal(activity.successors, successors)
    assert_equal(activity.delays, None)

    assert_equal(activity.num_modes, 1)
    assert_equal(activity.modes[0].duration, 2)
    assert_equal(activity.modes[0].demands, [0, 4, 8, 0])

    assert_equal(instance.num_projects, 1)
    assert_equal(instance.projects[0].num_activities, 18)


def test_instance_mmlib():
    """
    Tests an PSPLIB-style instance from MMLIB, which are formatted slightly
    different with header names.
    """
    instance = parse_psplib(relative("data/Jall1_1.mm"))

    capacities = [res.capacity for res in instance.resources]
    renewables = [res.renewable for res in instance.resources]

    assert_equal(instance.num_resources, 4)
    assert_equal(capacities, [33, 33, 247, 248])
    assert_equal(renewables, [True, True, False, False])

    assert_equal(instance.num_activities, 52)

    activity = instance.activities[1]  # second activity (jobnr. 2)
    successors = [50, 49, 47, 24, 22, 20, 19, 17, 16, 13]
    assert_equal(activity.successors, successors)
    assert_equal(activity.delays, None)

    assert_equal(activity.num_modes, 3)
    assert_equal(activity.modes[0].duration, 2)
    assert_equal(activity.modes[0].demands, [8, 8, 2, 8])
    assert_equal(activity.modes[1].duration, 3)
    assert_equal(activity.modes[1].demands, [5, 5, 2, 6])
    assert_equal(activity.modes[2].duration, 4)
    assert_equal(activity.modes[2].demands, [4, 5, 2, 6])

    assert_equal(instance.num_projects, 1)
    assert_equal(instance.projects[0].num_activities, 52)



from numpy.testing import assert_equal

from psplib import parse_rcpsp_max

from .utils import relative


def test_ubo10():
    """
    Tests that the instance ``UBO10_01.rcpsp`` is correctly parsed.
    """
    instance = parse_rcpsp_max(relative("data/UBO10_01.sch"))

    capacities = [res.capacity for res in instance.resources]
    renewables = [res.renewable for res in instance.resources]

    assert_equal(instance.num_resources, 5)
    assert_equal(capacities, [10, 10, 10, 10, 10])
    assert_equal(renewables, [True, True, True, True, True])

    assert_equal(instance.num_activities, 12)

    activity = instance.activities[2]  # third activity
    assert_equal(activity.successors, [4, 11, 7])
    assert_equal(activity.delays, [5, 9, 0])

    assert_equal(activity.num_modes, 1)
    assert_equal(activity.modes[0].demands, [10, 8, 0, 8, 10])
    assert_equal(activity.modes[0].duration, 9)

    assert_equal(instance.num_projects, 1)
    assert_equal(instance.projects[0].num_activities, 12)



from pathlib import Path
from typing import Union

from .ProjectInstance import Activity, Mode, Project, ProjectInstance, Resource


def parse_patterson(loc: Union[str, Path]) -> ProjectInstance:
    """
    Parses a Patterson-formatted instance. This format is used for pure
    resource-constrained project scheduling problem (RCPSP) instances.

    Parameters
    ----------
    loc
        The location of the instance.

    Returns
    -------
    ProjectInstance
        The parsed project instance.
    """
    with open(loc, "r") as fh:
        lines = iter(line.strip() for line in fh.readlines() if line.strip())

    num_activities, num_resources = map(int, next(lines).split())

    # Instances without resources do not have an availability line.
    capacities = list(map(int, next(lines).split())) if num_resources else []
    resources = [Resource(capacity=cap, renewable=True) for cap in capacities]

    # Most instances are not nicely formatted since a single activity data
    # may be split over multiple lines. The way to deal with this is to iterate
    # over all values instead of parsing line-by-line.
    values = iter([val for line in lines for val in line.split()])
    activities = []

    for _ in range(num_activities):
        duration = int(next(values))
        demands = [int(next(values)) for _ in range(num_resources)]
        num_successors = int(next(values))
        successors = [int(next(values)) - 1 for _ in range(num_successors)]
        activities.append(Activity([Mode(duration, demands)], successors))

    project = Project(list(range(num_activities)))  # only one project
    return ProjectInstance(resources, activities, [project])



import re
from pathlib import Path
from typing import Union

from .ProjectInstance import Activity, Mode, Project, ProjectInstance, Resource


def _find(lines: list[str], pattern: str) -> int:
    for idx, line in enumerate(lines):
        if pattern in line:
            return idx

    raise ValueError(f"Pattern '{pattern}' not found in lines.")


def parse_psplib(loc: Union[str, Path]) -> ProjectInstance:
    """
    Parses an PSPLIB-formatted instance from a file.

    Parameters
    ----------
    loc
        The location of the instance.

    Returns
    -------
    Instance
        The parsed instance.
    """
    with open(loc) as fh:
        lines = [line.strip() for line in fh.readlines() if line.strip()]

    prec_idx = _find(lines, "PRECEDENCE RELATIONS")
    req_idx = _find(lines, "REQUESTS/DURATIONS")
    avail_idx = _find(lines, "AVAILABILITIES")

    capacities = list(map(int, re.split(r"\s+", lines[avail_idx + 2])))
    renewable = [
        char == "R"
        for char in lines[avail_idx + 1].strip().split()
        if char in ["R", "N"]  # R: renewable, N: non-renewable
    ]
    resources = [
        Resource(capacity, is_renewable)
        for capacity, is_renewable in zip(capacities, renewable)
    ]
    num_resources = len(resources)

    mode_data = [
        list(map(int, re.split(r"\s+", line)))
        for line in lines[req_idx + 3 : avail_idx - 1]
    ]

    mode_idx = 0
    activities = []
    for line in lines[prec_idx + 2 : req_idx - 1]:
        _, num_modes, _, *jobs = list(map(int, re.split(r"\s+", line)))
        successors = [val - 1 for val in jobs if val]

        modes = []
        for idx in range(mode_idx, mode_idx + num_modes):
            # Mode data is parsed starting from the end of the line because
            # the data is ragged as some lines do not contain job indices.
            duration = mode_data[idx][-num_resources - 1]
            demands = mode_data[idx][-num_resources:]
            modes.append(Mode(duration, demands))
            mode_idx += 1

        activities.append(Activity(modes, successors))

    idcs = list(range(len(activities)))
    project = Project(idcs)  # only one project with all activities

    return ProjectInstance(resources, activities, [project])



from dataclasses import dataclass, field
from typing import Optional


@dataclass
class Resource:
    """
    Resource class.

    Parameters
    ----------
    capacity
        The available maximum capacity of the resource.
    renewable
        Whether the resource is renewable or not.
    """

    capacity: int
    renewable: bool


@dataclass
class Mode:
    """
    Mode class.

    Parameters
    ----------
    duration
        The duration of this processing mode.
    demands
        The resource demands (one per resource) of this processing mode.
    """

    duration: int
    demands: list[int]


@dataclass
class Activity:
    """
    Activity class.

    Parameters
    ----------
    modes
        The processing modes of this activity.
    successors
        The indices of successor activities.
    delays
        The delay for each successor activity. If delays are specified, then
        the length of this list must be equal to the length of `successors`.
        Delays are used for RCPSP/max instances, where the precedence
        relationship is defined as ``start(pred) + delay <= start(succ)``.
    optional
        Whether this activity is optional or not. Default ``False``.
    selection_groups
        The selection groups of this activity. If the current activity is
        scheduled, then for each group, exactly one activity must be scheduled.
        This is used for RCPSP-PS instances. Default is an empty list.
    name
        Optional name of the activity to identify this activity. This is
        helpful to map this activity back to the original problem instance.
    """

    modes: list[Mode]
    successors: list[int]
    delays: Optional[list[int]] = None
    optional: bool = False
    selection_groups: list[list[int]] = field(default_factory=list)
    name: str = ""

    def __post_init__(self):
        if self.delays and len(self.delays) != len(self.successors):
            raise ValueError("Length of successors and delays must be equal.")

    @property
    def num_modes(self):
        return len(self.modes)


@dataclass
class Project:
    """
    A project is a collection of activities that share a common release date
    and the project is considered finished when all activities are completed.

    Mainly used in multi-project instances. In regular project scheduling
    instances, there is only one project that contains all activities.

    Parameters
    ----------
    activities
        The activities indices that belong to this project.
    release_date
        The earliest start time of this project.
    """

    activities: list[int]
    release_date: int = 0

    @property
    def num_activities(self):
        return len(self.activities)


@dataclass
class ProjectInstance:
    """
    The project scheduling instance.
    """

    resources: list[Resource]
    activities: list[Activity]
    projects: list[Project]

    @property
    def num_resources(self):
        return len(self.resources)

    @property
    def num_activities(self):
        return len(self.activities)

    @property
    def num_projects(self):
        return len(self.projects)



from collections import defaultdict, deque
from dataclasses import dataclass
from itertools import chain
from pathlib import Path
from typing import Union

from .ProjectInstance import (
    Activity,
    Mode,
    Project,
    ProjectInstance,
    Resource,
)


@dataclass
class AlternativeSubgraph:
    """
    Represents a single alternative subgraph in an ASLIB instance.

    Parameters
    ----------
    branches
        A list of branches in the subgraph. Each branch is a list of activity
        indices that are part of the branch.
    """

    branches: list[list[int]]


def _parse_part_a(lines):
    """
    Part (a) of ASLIB instance is formatted as Patterson instance.
    """
    num_activities, num_resources = map(int, next(lines).split())

    # Instances without resources do not have an availability line.
    capacities = list(map(int, next(lines).split())) if num_resources else []
    resources = [Resource(capacity=cap, renewable=True) for cap in capacities]

    activities = []
    for _ in range(num_activities):
        values = map(int, next(lines).split())
        duration = int(next(values))
        demands = [int(next(values)) for _ in range(num_resources)]
        num_successors = int(next(values))
        successors = [int(next(values)) - 1 for _ in range(num_successors)]
        activities.append(Activity([Mode(duration, demands)], successors))

    return resources, activities


def _parse_part_b(lines) -> list[AlternativeSubgraph]:
    """
    Part (b) of ASLIB instance results in alternative subgraphs.
    """
    pct_flex, pct_nested, pct_linked = map(float, next(lines).split())
    num_subgraphs = int(next(lines))
    total_branches = 1  # first branch is always the dummy branch
    subgraphs = []

    for _ in range(num_subgraphs):
        num_branches, *branch_idcs = map(int, next(lines).split())
        total_branches += num_branches
        branch_idcs = [idx - 1 for idx in branch_idcs]
        subgraphs.append(branch_idcs)

    branches: list[list[int]] = [[] for _ in range(total_branches)]
    for activity, line in enumerate(lines):
        num_braches, *branch_idcs = map(int, line.split())
        for idx in branch_idcs:
            branches[idx - 1].append(activity)

    # Return the alternative subgraphs, with the first subgraph containing the
    # fixed activities (an activity belongs to node branch 0 if it is fixed).
    result = [AlternativeSubgraph([branches[0]])]
    result += [
        AlternativeSubgraph([branches[idx] for idx in branch_idcs])
        for branch_idcs in subgraphs
    ]

    return result


def parse_aslib(loc: Union[str, Path]) -> ProjectInstance:
    """
    Parses an ASLIB-formatted instance from a file. This format is used for
    RCPSP instances with alternative subgraphs.

    Note
    ----
    This function parses files that combine both "a" and "b" part files from
    the ASLIB instance. You have to manually create such instances first!

    Parameters
    ----------
    loc
        The location of the instance.

    Returns
    -------
    ProjectInstance
        The parsed project instance.
    """
    with open(loc, "r") as fh:
        lines = iter(line.strip() for line in fh.readlines() if line.strip())

    resources, activities = _parse_part_a(lines)
    subgraphs = _parse_part_b(lines)

    # With the already parsed activities and alternative subgraph data,
    # we add optional and selections groups data to the activities.
    activities = _make_optional_activities(activities, subgraphs)

    project = Project(list(range(len(activities))))
    return ProjectInstance(resources, activities, [project])


class DiGraph:
    """
    Simple directed graph implementation to replace networkx.DiGraph.
    """

    def __init__(self):
        self.adj: dict[int, list[int]] = defaultdict(list)
        self.nodes = set()

    def add_node(self, node: int):
        self.nodes.add(node)

    def add_edge(self, u: int, v: int):
        self.adj[u].append(v)
        self.nodes.add(u)
        self.nodes.add(v)

    def topological_sort(self) -> list[int]:
        """
        Returns a topological ordering of the graph's nodes.
        """
        in_degree = {node: 0 for node in self.nodes}
        for u in self.adj:
            for v in self.adj[u]:
                in_degree[v] += 1

        queue = deque(node for node in self.nodes if in_degree[node] == 0)
        order = []
        while queue:
            node = queue.popleft()
            order.append(node)
            for neighbor in self.adj[node]:
                in_degree[neighbor] -= 1
                if in_degree[neighbor] == 0:
                    queue.append(neighbor)

        return order

    @classmethod
    def from_activities(cls, activities: list[Activity]):
        G = cls()
        for idx, activity in enumerate(activities):
            G.add_node(idx)
            for succ in activity.successors:
                G.add_edge(idx, succ)

        return G


def _make_optional_activities(
    activities: list[Activity],
    subgraphs: list[AlternativeSubgraph],
) -> list[Activity]:
    """
    Adds optional and selection group data to activities based on alternative
    subgraph data. Because the activity graphs are directed acylcic, we can
    transform subgraphs into selection groups.
    """
    G = DiGraph.from_activities(activities)
    order = G.topological_sort()

    is_fixed = subgraphs[0].branches[0]  # first subgraph contains fixed nodes
    alternatives = subgraphs[1:]
    all_branching = []  # idcs of branching activities
    groups = defaultdict(list)

    for subgraph in alternatives:
        # The branching activities are the lowest-indexed activities in each
        # branch. This works because we have a directed acyclic graph.
        branching = [min(b, key=order.index) for b in subgraph.branches]
        arcs = [(u, v) for u in G.adj for v in G.adj[u] if v in branching]

        # The principal activity is the sole activity that goes to the
        # branching activities.
        nodes = {u for (u, _) in arcs}
        assert len(nodes) == 1  # should be only 1 principal activity
        principal = nodes.pop()

        all_branching.extend(branching)
        groups[principal].append(branching)

    # For all remaining edges, we add another unit selection group if
    # v is not a branching activity. Edges with v as a branching activity
    # are already covered by the groups above.
    for u in G.adj:
        for v in G.adj[u]:
            if v not in all_branching:
                groups[u].append([v])

    # Create new activities with optional and selection group data.
    new = []
    for idx, activity in enumerate(activities):
        activity = Activity(
            modes=activity.modes,
            successors=activity.successors,
            optional=idx not in is_fixed,
            selection_groups=groups[idx],
        )
        new.append(activity)

    for activity in new:
        # Check: In RCPSP-AS, timing successors are also selection successors.
        select_succ = list(chain(*activity.selection_groups))
        assert sorted(select_succ) == sorted(activity.successors)

    return new



import pytest

from psplib import parse
from tests.utils import relative


@pytest.mark.parametrize(
    "loc, instance_format",
    [
        ("data/Jall1_1.mm", "psplib"),
        ("data/RG300_1.rcp", "patterson"),
        ("data/MPLIB1_Set1_0.rcmp", "mplib"),
        ("data/UBO10_01.sch", "rcpsp_max"),
        ("data/rcpsp_ps.txt", "rcpsp_ps"),
    ],
)
def test_parse(loc, instance_format):
    """
    Checks that the instances are successfully parsed.
    """
    parse(relative(loc), instance_format)



from numpy.testing import assert_equal

from psplib import parse_patterson

from .utils import relative


def test_rg300():
    """
    Tests that the instance ``RG300_1.rcp`` is correctly parsed.
    """
    instance = parse_patterson(relative("data/RG300_1.rcp"))

    capacities = [res.capacity for res in instance.resources]
    renewables = [res.renewable for res in instance.resources]

    assert_equal(instance.num_resources, 4)
    assert_equal(capacities, [10, 10, 10, 10])
    assert_equal(renewables, [True, True, True, True])

    assert_equal(instance.num_activities, 302)

    activity = instance.activities[1]  # second activity

    successors = (
        "59, 79, 88, 95, 104, 113, 125, 128, 138, 141, 145, 146, 151, "
        "153, 155, 156, 159, 161, 163, 169, 171, 174, 176, 178, 180, "
        "201, 204, 212, 285, 287, 289, 291, 292"
    )
    successors = list(map(int, successors.split(", ")))
    assert_equal(activity.successors, successors)
    assert_equal(activity.delays, None)

    assert_equal(activity.num_modes, 1)
    assert_equal(activity.modes[0].demands, [0, 1, 0, 0])
    assert_equal(activity.modes[0].duration, 3)

    assert_equal(instance.num_projects, 1)
    assert_equal(instance.projects[0].num_activities, 302)



from numpy.testing import assert_equal

from psplib import parse_mplib

from .utils import relative


def test_mplib_set1():
    """
    Tests that the instance ``MPLIB1_Set1_0.rcmp`` is correctly parsed.
    """
    instance = parse_mplib(relative("data/MPLIB1_Set1_0.rcmp"))

    capacities = [res.capacity for res in instance.resources]
    renewables = [res.renewable for res in instance.resources]

    assert_equal(instance.num_resources, 4)
    assert_equal(capacities, [56, 56, 56, 56])
    assert_equal(renewables, [True, True, True, True])

    assert_equal(instance.num_activities, 6 * 62)

    activity = instance.activities[0]

    # Successors are 1:2, 1:3, 1:4 -> activity indices 1, 2, 3
    assert_equal(activity.successors, [1, 2, 3])
    assert_equal(activity.delays, None)

    assert_equal(activity.num_modes, 1)
    assert_equal(activity.modes[0].demands, [0, 0, 0, 0])
    assert_equal(activity.modes[0].duration, 0)

    assert_equal(instance.num_projects, 6)
    for project in instance.projects:
        assert_equal(project.num_activities, 62)
        assert_equal(project.release_date, 0)


def test_mplib_set2():
    """
    Tests that the instance ``MPLIB2_Set1_0.rcmp`` is correctly parsed.
    """
    instance = parse_mplib(relative("data/MPLIB2_Set1_0.rcmp"))

    capacities = [res.capacity for res in instance.resources]
    renewables = [res.renewable for res in instance.resources]

    assert_equal(instance.num_resources, 5)
    assert_equal(capacities, [48, 48, 46, 50, 48])
    assert_equal(renewables, [True, True, True, True, True])

    assert_equal(instance.num_activities, 10 * 52)

    activity = instance.activities[-51]  # second activity of last project

    # Successors are 10:14 10:13 10:12 10:9.
    successors = [(9 * 52) + idx - 1 for idx in [14, 13, 12, 9]]
    assert_equal(activity.successors, successors)
    assert_equal(activity.delays, None)

    assert_equal(activity.num_modes, 1)
    assert_equal(activity.modes[0].demands, [8, 4, 3, 5, 1])
    assert_equal(activity.modes[0].duration, 7)

    assert_equal(instance.num_projects, 10)
    for project in instance.projects:
        assert_equal(project.num_activities, 52)
        assert_equal(project.release_date, 0)



from pathlib import Path
from typing import Union

from .parse_aslib import parse_aslib
from .parse_mplib import parse_mplib
from .parse_patterson import parse_patterson
from .parse_psplib import parse_psplib
from .parse_rcpsp_max import parse_rcpsp_max
from .parse_rcpsp_ps import parse_rcpsp_ps
from .ProjectInstance import ProjectInstance


def parse(
    loc: Union[str, Path],
    instance_format: str = "psplib",
) -> ProjectInstance:
    """
    Parses a project instance from a file location.

    Parameters
    ----------
    loc
        The location of the instance.
    instance_format
        The format of the instance.

    Returns
    -------
    ProjectInstance
        The parsed project instance.
    """
    if instance_format == "psplib":
        return parse_psplib(loc)
    elif instance_format == "patterson":
        return parse_patterson(loc)
    elif instance_format == "rcpsp_max":
        return parse_rcpsp_max(loc)
    elif instance_format == "mplib":
        return parse_mplib(loc)
    elif instance_format == "rcpsp_ps":
        return parse_rcpsp_ps(loc)
    elif instance_format == "aslib":
        return parse_aslib(loc)

    raise ValueError(f"Unknown instance format: {instance_format}")



from numpy.testing import assert_equal

from psplib import parse_aslib

from .utils import relative


def test_aslib0_0():
    """
    Tests that the instance ``aslib0_0.rcp`` is correctly parsed.
    """
    instance = parse_aslib(relative("data/aslib0_0.rcp"))
    assert_equal(instance.num_resources, 5)

    capacities = [res.capacity for res in instance.resources]
    renewables = [res.renewable for res in instance.resources]

    assert_equal(capacities, [10, 10, 10, 10, 10])
    assert_equal(renewables, [True, True, True, True, True])

    assert_equal(instance.num_activities, 122)

    activity = instance.activities[0]  # source
    assert_equal(activity.successors, [1, 13, 25, 37, 49])
    assert_equal(activity.optional, False)  # source always present
    assert_equal(activity.selection_groups, [[1, 13, 25, 37, 49]])

    assert_equal(activity.num_modes, 1)
    assert_equal(activity.modes[0].demands, [0, 0, 0, 0, 0])
    assert_equal(activity.modes[0].duration, 0)

    activity = instance.activities[2]
    assert_equal(activity.successors, [9, 8])
    assert_equal(activity.optional, True)
    assert_equal(activity.selection_groups, [[9], [8]])

    assert_equal(activity.num_modes, 1)
    assert_equal(activity.modes[0].demands, [0, 0, 1, 0, 0])
    assert_equal(activity.modes[0].duration, 1)

    assert_equal(instance.num_projects, 1)
    assert_equal(instance.projects[0].num_activities, 122)



from pathlib import Path
from typing import Union

from .ProjectInstance import Activity, Mode, Project, ProjectInstance, Resource


def parse_rcpsp_max(loc: Union[str, Path]) -> ProjectInstance:
    """
    Parses the RCPSP/max instance format.

    Parameters
    ----------
    loc
        The location of the instance.

    Returns
    -------
    ProjectInstance
        The parsed project instance.
    """

    with open(loc, "r") as fh:
        lines = iter(line.strip() for line in fh.readlines() if line.strip())

    num_activities, num_renewables, *_ = map(int, next(lines).split())
    num_activities += 2  # source and target
    activities = []

    succ_lines = [next(lines).split() for _ in range(num_activities)]
    activity_lines = [next(lines).split() for _ in range(num_activities)]

    for idx in range(num_activities):
        _, _, num_successors, *succ = succ_lines[idx]
        successors = list(map(int, succ[: int(num_successors)]))
        delays = [int(val.strip("[]")) for val in succ[int(num_successors) :]]

        _, _, duration, *demands = map(int, activity_lines[idx])

        activity = Activity([Mode(duration, demands)], successors, delays)
        activities.append(activity)

    capacities = map(int, next(lines).split())
    resources = [Resource(capacity, renewable=True) for capacity in capacities]
    projects = [Project(list(range(num_activities)))]

    return ProjectInstance(resources, activities, projects)



from .parse import parse as parse
from .parse_aslib import parse_aslib as parse_aslib
from .parse_mplib import parse_mplib as parse_mplib
from .parse_patterson import parse_patterson as parse_patterson
from .parse_psplib import parse_psplib as parse_psplib
from .parse_rcpsp_max import parse_rcpsp_max as parse_rcpsp_max
from .parse_rcpsp_ps import parse_rcpsp_ps as parse_rcpsp_ps
from .ProjectInstance import ProjectInstance as ProjectInstance



from numpy.testing import assert_equal

from psplib import parse_rcpsp_ps

from .utils import relative


def test_rcpsp_ps():
    """
    Tests that the instance ``rcpsp_ps.txt`` is correctly parsed.
    """
    instance = parse_rcpsp_ps(relative("data/rcpsp_ps.txt"))

    assert_equal(instance.num_resources, 4)

    capacities = [res.capacity for res in instance.resources]
    renewables = [res.renewable for res in instance.resources]

    assert_equal(capacities, [10, 10, 10, 10])
    assert_equal(renewables, [True, True, True, True])

    assert_equal(instance.num_activities, 136)

    activity = instance.activities[0]  # first activity (source)
    assert_equal(activity.successors, [1, 2])
    assert_equal(activity.optional, False)  # source always present
    assert_equal(activity.selection_groups, [[1, 2]])

    assert_equal(activity.num_modes, 1)
    assert_equal(activity.modes[0].demands, [0, 0, 0, 0])
    assert_equal(activity.modes[0].duration, 0)

    activity = instance.activities[3]  # fourth activity
    assert_equal(activity.successors, [5])
    assert_equal(activity.selection_groups, [[5]])

    assert_equal(activity.num_modes, 1)
    assert_equal(activity.modes[0].demands, [0, 0, 1, 1])
    assert_equal(activity.modes[0].duration, 9)

    assert_equal(instance.num_projects, 1)
    assert_equal(instance.projects[0].num_activities, 136)
